
program Quatre;

{ Puissance 4 en mode console
  R. Chastain septembre 2012
  
  Compilation teste :
  - Turbo Pascal 7.0
  - Virtual Pascal 2.1
  - Free Pascal 2.6.0
  - Delphi 7
  - Delphi XE2
}

{$IfDef VPascal}
{&PmType Vio}
{$Else}
{$IfNDef Ver70}
{$AppType Console}
{$EndIf}
{$EndIf}

{$IfNDef Ver230}
{$Define UsesCrt}
{$IfDef Ver150}
uses Crt in 'C:\Program files\Borland\Delphi7\Source\Crt\crt.pas';
{$Else}
uses Crt;
{$EndIf}
{$EndIf}

{$IfNDef Ver70}
{$R Icone}
{$EndIf}

type
  tEtatJeu = (marche, arret);
  tPion = (neant, blanc, noir);
  tJoueur = (utilisateur, machine);

  tGrille = array[0..5, 0..6] of tPion;

  tJeu = object
    grille: tGrille;
    etat: tEtatJeu;
    trait: tPion;
    joueurs: array[tPion] of tJoueur;
    procedure Commence;
    procedure Represente;
    procedure Joue;
    function CoupUtilisateur: integer;
    procedure Enregistre(coup: integer);
    function GrillePleine: boolean;
    function GrilleGagnante: boolean;
    function MeilleurCoup: integer;
  end;

var
  jeu: tJeu;
  saisie: string;

procedure tJeu.Commence;
var
  x, y: integer;
begin
  with self do
  begin
    etat := marche;
    for y := 0 to 5 do for x := 0 to 6 do
      grille[y, x] := neant;
    trait := blanc;
    joueurs[blanc] := utilisateur;
    joueurs[noir] := machine;
  end;
end;

function Score(const cG: tGrille; cP: tPion): integer; forward;

procedure tJeu.Represente;
var
  x, y: integer;
  lettre: char;
begin
  WriteLn;
  for y := 0 to 5 do
  begin
    WriteLn('+---+---+---+---+---+---+---+');
    for x := 0 to 6 do
    begin
      Write('| ');
      case self.grille[y, x] of
        neant: lettre := ' ';
        blanc: begin
                 lettre := 'x';
                 {$IfDef UsesCrt}
                 TextColor(Yellow);
                 {$EndIf}
               end;
        noir: begin
                lettre := 'o';
                {$IfDef UsesCrt}
                TextColor(LightRed);
                {$EndIf}
              end;
      else
        lettre := 'e';
      end;
      Write(lettre);
      {$IfDef UsesCrt}
      TextColor(LightGray);
      {$EndIf}
      if x < 6 then
        Write(' ')
      else
        Write(' |');
    end;
    WriteLn;
  end;
  WriteLn('+---+---+---+---+---+---+---+');
  WriteLn('| a | b | c | d | e | f | g |');
  WriteLn('+---+---+---+---+---+---+---+');
  WriteLn;
 {WriteLn(Score(self.grille, blanc));
  WriteLn(Score(self.grille, noir));
  WriteLn;}
end;

procedure tJeu.Joue;
var
  mc: integer;
begin
  with self do
  begin
    if joueurs[trait] = utilisateur then
      Enregistre(CoupUtilisateur)
    else
    begin
      mc := MeilleurCoup;
      Enregistre(mc);
      WriteLn('!' + chr(mc+97));
    end;
    Represente;
  end;
end;

function tJeu.CoupUtilisateur: integer;
const
  col_maj: string = 'ABCDEFG';
  col_min: string = 'abcdefg';
var
  temp: integer;
begin
  Write('?');
  ReadLn(saisie);
  if Length(saisie) <> 1 then
    Halt;
  temp := Pos(saisie, col_maj);
  if temp = 0 then
    temp := Pos(saisie, col_min);
  if temp = 0 then
    temp := 7
  else
    Dec(temp);
  CoupUtilisateur := temp;
end;

procedure tJeu.Enregistre(coup: Integer);
var
  y: integer;
begin
  if coup = 7 then
    Halt;
  with self do
  begin
    if grille[0, coup] = neant then
    begin
      y := 0;
      while (y < 5) and (grille[y + 1, coup] = neant) do
        Inc(y);
      grille[y, coup] := trait;

      if GrillePleine or GrilleGagnante then
      begin
        etat := arret;
        Exit;
      end;

      if trait = blanc then trait := noir
      else
        trait := blanc;
    end;
  end;
end;

function tJeu.GrillePleine: boolean;
var
  x, y: integer;
  temp: boolean;
begin
  temp := true;
  for y := 0 to 5 do for x := 0 to 6 do if self.grille[y, x] = neant then
  begin
    temp := false;
    GrillePleine := temp;
    Exit;
  end;
  GrillePleine := temp;
end;

function tJeu.GrilleGagnante: boolean;
var temp: boolean;
begin
  temp := false;
  with self do if Score(grille, trait) >= 4 then
    temp := true;
  GrilleGagnante := temp;
end;

function Score(const cG: tGrille; cP: tPion): integer;
const
  coord: array[0..11, 0..1] of integer = ((0,3),
                                          (0,4),
                                          (0,5),
                                          (1,5),
                                          (2,5),
                                          (3,5),
                                          (6,3),
                                          (6,4),
                                          (6,5),
                                          (5,5),
                                          (4,5),
                                          (3,5));
var
  n, nmax, i, x, y: integer;
begin
  nmax := 0;
  
  for i := 0 to 5 do
  begin
    x := coord[i, 0];
    y := coord[i, 1];
    n := 0;
    while (x >= 0) and (x <= 6) and (y >= 0) and (y <= 5) do
    begin
      if cG[y, x] = cP then inc(n) else n := 0;
      if n > nmax then nmax := n;
      inc(x);
      dec(y);
    end;
  end;
  
  for i := 6 to 11 do
  begin
    x := coord[i, 0];
    y := coord[i, 1];
    n := 0;
    while (x >= 0) and (x <= 6) and (y >= 0) and (y <= 5) do
    begin
      if cG[y, x] = cP then inc(n) else n := 0;
      if n > nmax then nmax := n;
      dec(x);
      dec(y);
    end;
  end;
  
  for x := 0 to 6 do
  begin
    n := 0;
    for y := 5 downTo 0 do
    begin
      if cG[y, x] = cP then inc(n) else n := 0;
      if n > nmax then nmax := n;
    end;
  end;
  
  for y := 5 downTo 0 do
  begin
    n := 0;
    for x := 0 to 6 do
    begin
      if cG[y, x] = cP then inc(n) else n := 0;
      if n > nmax then nmax := n;
    end;
  end;
  
  Score := nmax;
end;

procedure Copie(const cG: tGrille; var vG: tGrille);
var
  x, y: integer;
begin
  for x := 0 to 6 do for y := 0 to 5 do vG[y, x] := cG[y, x];
end;

function Ajoute(var vG: tGrille; const cX: integer; const cP: tPion): boolean;
var
  y: integer;
  temp: boolean;
begin
  y := 0;
  if vG[y, cX] = neant then
  begin
    temp := true;
    while (y < 5) and (vG[y + 1, cX] = neant) do Inc(y);
    vG[y, cX] := cP;
  end else
    temp := false;
  Ajoute := temp;
end;

function tJeu.MeilleurCoup: integer;
var
  g1, g2: tGrille;
  x1, x2: integer;
  note: array[0..6] of integer;
  notemax: integer;
  max_adv: integer;
  sg1: integer;
  sg2: array[0..6] of integer;
  
  function pion_adv(p: tPion): tPion;
  begin
    if p = blanc
    then pion_adv := noir
    else pion_adv := blanc;
  end;
  
begin
  notemax := - 100;
  
  for x1 := 0 to 6 do
  begin
    note[x1] := 0;
    Copie(self.grille, g1);
    if not Ajoute(g1, x1, self.trait) then note[x1] := note[x1] - 10000;
    
    sg1 := Score(g1, self.trait);
    if sg1 >= 4 then note[x1] := note[x1] + 1000;
    
    max_adv := 0;
    for x2 := 0 to 6 do
    begin
      Copie(g1, g2);
      Ajoute(g2, x2, pion_adv(self.trait));
      sg2[x2] := Score(g2, pion_adv(self.trait));
      if sg2[x2] > max_adv then max_adv := sg2[x2];
    end;
    
    if max_adv >= 4 then note[x1] := note[x1] - 100
    else
      for x2 := 0 to 6 do
        if sg2[x2] = max_adv then note[x1] := note[x1] - max_adv;
    
    note[x1] := note[x1] - Abs(x1-3);
    if note[x1] > notemax then
      notemax := note[x1];
  end;
  
  x1 := 0;
  while note[x1] < notemax do
    Inc(x1);
  
  MeilleurCoup := x1;
end;

label
  start;

begin
  { --- Puissance 4 -------------------------------------------------------- }

  start:

  jeu.Commence;
  jeu.Represente;

  while jeu.etat = marche do jeu.Joue;

  WriteLn('Voulez-vous commencer une autre partie (o/n)?');
  ReadLn(saisie);

  if saisie = 'o' then
    goto start;

  { --- Fin Puissance 4 ---------------------------------------------------- }
end.
